"""Implementation of Equation class for representing conservation laws"""

import sympy as sp


class Equation(sp.core.relational.Equality):
    """Sympy equation where we overload addition/subtraction to apply those operations to the RHS, for summing rate
    equations"""

    def get_summand(self, other):
        """Value-check the operand and return the quantity to be summed in the operation: the expression itself if an expression, or the RHS"""
        if isinstance(other, sp.core.relational.Equality):
            if self.lhs != other.lhs:
                raise ValueError(
                    "Tried to sum incompatible equations. Equation summation only defined for differential equations with the same LHS."
                )
            else:
                return other.rhs
        elif isinstance(other, sp.logic.boolalg.BooleanAtom):
            return 0
        else:
            return other

    def __add__(self, other):
        summand = self.get_summand(other)
        return Equation(self.lhs, self.rhs + summand)

    def __sub__(self, other):
        summand = self.get_summand(other)
        return Equation(self.lhs, self.rhs - summand)

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        self = self + other
        return self

    def __isub__(self, other):
        self = self - other
        return self



import numpy as np
import jax, jax.numpy as jnp
import jaco
from jaco.numerics.solvers import newton_rootsolve


def test_newton_rootsolve(N=10**5):
    """Test: solve the system x^p = a for various choices of p and a and check solution

    NOTE: this newton iteration does not converge in general, so we will not catch all possible bugs that might return
    nonfinite values...
    """
    import numpy as np

    p = 0.1 + np.random.rand(N) * 10
    a = 0.1 + np.random.rand(N)
    params = jnp.c_[p, a]
    guess = jnp.ones(N)

    exact = np.atleast_2d(a ** (1.0 / p)).T

    def func(x, *params):
        return x ** params[0] - params[1]

    func = jax.jit(func)

    sol = newton_rootsolve(func, guess, params, nonnegative=True)
    converged = jnp.all(jnp.isfinite(sol), axis=1)
    assert converged.sum() > 0.9 * N
    assert jnp.all(jnp.isclose(sol[converged], exact[converged], rtol=1e-3, atol=0))


if __name__ == "__main__":
    test_newton_rootsolve()



import sympy as sp
from ..process import Process
from .nbody_process import NBodyProcess
from ..symbols import T, T5, n_
from ..data import SolarAbundances

# put analytic fits for cooling efficiencies
line_cooling_coeffs = {
    "H": {"e-": 7.5e-19 * sp.exp(-118348 / T) / (1 + sp.sqrt(T5))},  # 1996ApJS..105...19K
    "He+": {"e-": 5.54e-17 * T**-0.397 * sp.exp(-473638 / T) / (1 + sp.sqrt(T5))},  # 1996ApJS..105...19K
    "C+": {  # 2023MNRAS.519.3154H
        "e-": 1e-27 * 4890 / sp.sqrt(T) * sp.exp(-91.211 / T) / SolarAbundances.x("C"),
        "H": 1e-27 * 0.47 * T**0.15 * sp.exp(-91.211 / T) / SolarAbundances.x("C"),
    },
}


def LineCoolingSimple(emitter: str, collider=None) -> NBodyProcess:
    """Returns a 2-body process representing cooling via excitations from collisions of given pair of species

    This is the simple approximation where everything is well below critical density and no ambient radiation field.
    eventually would like to have a class that considers collisions from all available colliders, given just the
    energies, deexcitation coefficients, temperature, and statistical weights...

    Parameters
    ----------
    emitter: str
        Emitting excited species
    collider: str, optional
        Exciting colliding species. If None, will look up all known

    Returns
    -------
    An NBodyProcess instance whose heat attribute is the line cooling process's cooling rate in erg cm^-3
    """

    if emitter not in line_cooling_coeffs:
        raise NotImplementedError(f"Line cooling not implemented for {emitter}")

    coeffs = line_cooling_coeffs[emitter]

    if collider is None:  # if we haven't specified a collider, just take all of them and return the sum
        p = [LineCoolingSimple(emitter, c) for c in coeffs]
        return sum(p)  # type: ignore # have to put a 0-process in here as start variable or it will try to add 0 + process

    process = NBodyProcess({emitter, collider})
    if collider not in line_cooling_coeffs[emitter]:
        raise NotImplementedError(f"Excitation by collisions with {collider} not implemented for {emitter}")

    process.heat_rate_coefficient = -line_cooling_coeffs[emitter][collider]
    process.name = f"{emitter}-{collider} Line Cooling"

    return process


# def LineCooling(emitter: str)



from jaco.processes import CollisionalIonization, GasPhaseRecombination
import numpy as np
from matplotlib import pyplot as plt


def test_CIE():
    """Solve for CIE of a H-He mixture and check that the He+ abundance looks reasonable"""
    chianti_data = np.load("tests/chianti_He_abundances.npy")
    N = len(chianti_data)
    Tmin, Tmax = chianti_data[:, 0].min(), chianti_data[:, 0].max()

    processes = [CollisionalIonization(s) for s in ("H", "He", "He+")] + [
        GasPhaseRecombination(i) for i in ("H+", "He+", "He++")
    ]
    system = sum(processes)

    Tgrid = np.logspace(np.log10(Tmin), np.log10(Tmax), N)
    ngrid = np.ones_like(Tgrid)

    knowns = {"T": Tgrid, "n_Htot": ngrid}

    guesses = {"H": 0.5 * np.ones_like(Tgrid), "He": 1e-5 * np.ones_like(Tgrid), "He+": 1e-5 * np.ones_like(Tgrid)}
    sol = system.solve(knowns, guesses, tol=1e-3)
    Hep_frac = sol["He+"] / (sol["He"] + sol["He+"] + sol["He++"])
    assert (
        np.abs(Hep_frac - chianti_data[:, 2]).max() < 0.2
    )  # just a loose test because the codes don't use the same rates


def generate_CIE_testdata():
    import ChiantiPy.core as ch

    Tgrid = np.logspace(3, 6, 10**4)
    he = ch.ioneq(2)
    he.load()
    he.calculate(Tgrid)
    np.save("chianti_He_abundances.npy", np.c_[Tgrid, he.Ioneq])



"""Various convenience routines used throughout the package"""

from string import digits


def species_charge(species: str) -> int:
    """Returns the charge number of a species from its name"""
    if species[-1] not in ("-", "+"):
        return 0
    if "++" in species:
        return 2
    elif "--" in species:
        return -2

    base = base_species(species)
    suffix = species.split(base)[-1]
    if suffix == "+":
        return 1
    if suffix == "-":
        return -1
    elif "+" in suffix:
        return int(suffix.rstrip("+"))
    else:
        return -int(suffix.rstrip("-"))


def is_an_ion(species: str) -> bool:
    return species_charge(species) != 0 and species != "e-"


def base_species(species: str) -> str:
    """Removes the charge suffix from a species"""
    base = species.rstrip(digits + "-+")
    return base


def charge_suffix(charge: int) -> str:
    """Returns the suffix for an input charge number"""
    match charge:
        case 0:
            return ""
        case 1:
            return "+"
        case 2:
            return "++"
        case -1:
            return "-"
        case -2:
            return "--"
        case _:
            if charge < -2:
                return str(abs(charge)) + "-"
            else:
                return str(abs(charge)) + "+"


def ionize(species: str) -> str:
    """Returns the symbol of the species produced by removing an electron from the input species"""
    charge = species_charge(species)
    return base_species(species) + charge_suffix(charge + 1)


def recombine(species: str) -> str:
    """Returns the symbol of the species produced by adding an electron to the input species"""
    charge = species_charge(species)
    return base_species(species) + charge_suffix(charge - 1)



"""Definition of various sympy symbols used throughout the module"""

import sympy as sp
from typing import Union
import sympy as sp
from astropy.constants import k_B, m_p
from astropy import units as u

T = sp.Symbol("T")  # temperature
T5 = T / 10**5
T6 = T / 10**6
T3 = T / 10**3
T4 = T / 10**4
c_s = sp.Symbol("c_s")  # sound speed
G = sp.Symbol("G")  # gravitational constant
ρ = sp.Symbol("ρ")  # total mass density
n_e = sp.Symbol("n_e-")  # electron number density
z = sp.Symbol("z")  # cosmological redshift
t = sp.Symbol("t")  # time
dt = sp.Symbol("Δt")
n_Htot = sp.Symbol("n_Htot")


boltzmann_cgs = k_B.to(u.erg / u.K).value
protonmass_cgs = m_p.to(u.g).value
# write down internal energy density in terms of number densities - this defines the EOS


def d_dt(species: Union[str, sp.core.symbol.Symbol]):
    if isinstance(species, str):
        return sp.diff(sp.Function(n_(species))(t), t)
    else:
        return sp.diff(sp.Function(species)(t), t)


def n_(species: str):
    match species:
        case "heat":
            return sp.Symbol(f"⍴u")
        case _:
            return sp.Symbol(f"n_{species}")


egy_density = boltzmann_cgs * T * (1.5 * (n_("e-") + n_("H") + n_("H+") + n_("He") + n_("He+") + n_("He++")))
rho = protonmass_cgs * (n_("H") + n_("H+") + 4 * (n_("He") + n_("He+") + n_("He++")))
internal_energy = sp.factor(egy_density / rho)


def x_(species: str):
    return sp.Symbol(f"x_{species}")


def BDF(species):
    if species in ("T", "u"):  # this is the heat equation
        return rho * (internal_energy - sp.Symbol("u_0")) / dt
    else:
        return (n_(species) - sp.Symbol(str(n_(species)) + "_0")) / dt



"""Implementation of recombination process"""

from ..process import Process
from .nbody_process import NBodyProcess
from ..misc import recombine
from ..symbols import T, T3, T6
from .ionization import ionization_energy
import sympy as sp


class Recombination(NBodyProcess):
    """
    Class describing a recombination process: ion + e- -> recombined species + hν

    Implements method for setting the chemistry network terms

    Parameters
    ----------
    ion: str
        Ionic species being recombined
    """

    def __init__(self, ion: str):
        self.ion = ion
        self.recombined_species = recombine(ion)
        self.colliding_species = {ion, "e-"}
        super().__init__(self.colliding_species)
        self.ionization_energy = ionization_energy(self.recombined_species)
        self.__rate_coefficient = 0
        self.heat_rate_coefficient = 0

    @property
    def rate_coefficient(self):
        """Returns the rate coefficient of the recombination process"""
        return self.__rate_coefficient

    @rate_coefficient.setter
    def rate_coefficient(self, value):
        """Ensures that the network is always updated when we update the rate coefficient"""
        self.__rate_coefficient = value
        self.update_network()

    def update_network(self):
        """Sets up rate terms in the associated chemistry network for each ion involved"""
        if self.rate is None:
            return
        self.network[self.ion] -= self.rate
        self.network[self.recombined_species] += self.rate
        self.network["e-"] -= self.rate


def GasPhaseRecombination(ion=None) -> Recombination:
    """Return a recombination process representing gas-phase (e.g. radiative) recombination

    Parameters
    ----------
    ion: str, optional
        Ionic species getting recombined. If None, function will return a composite process of all gas-phase recombination
        processes with known rates.

    Returns
    -------
    process: Recombination
        `Process` instance describing the gas-phase recombination process
    """
    if ion is None:
        return sum([GasPhaseRecombination(s) for s in gasphase_recombination_rates], Process())

    process = Recombination(ion)
    process.name = f"Gas-phase recombination of {ion}"

    if ion not in gasphase_recombination_rates:
        raise NotImplementedError(f"{ion} does not have an available gas-phase recombination coefficient.")
    process.rate_coefficient = gasphase_recombination_rates[ion]
    process.heat_rate_coefficient = -gasphase_recombination_cooling[ion]
    return process


def hydrogenic_recombination_rate(Z):
    """Verner & Ferland 1996"""
    return (
        Z
        * 7.982e-11
        / (
            sp.sqrt(T / 3.148 / Z**2)
            * sp.Pow((1.0 + sp.sqrt(T / 3.148 / Z**2)), 0.252)
            * sp.Pow((1.0 + sp.sqrt(T / 7.036e5 / Z**2)), 1.748)
        )
    )


# All fits below are from Verner & Ferland 1996
gasphase_recombination_rates = {
    "H+": hydrogenic_recombination_rate(1),
    "He+": 9.356e-10
    / (
        sp.sqrt(T / 4.266e-2)
        * sp.Pow((1.0 + sp.sqrt(T / 4.266e-2)), 0.2108)
        * sp.Pow((1.0 + sp.sqrt(T / 3.676e7)), 1.7892)
    )
    + 1.9e-3 * T**-1.5 * sp.exp(-4.7e5 / T) * (1 + 0.3 * sp.exp(-9.4e4 / T)),
    "He++": hydrogenic_recombination_rate(2),
}
# an electron—ion pair removes the mean kinetic energy during recombination
# To a good approximation, the mean energy lost by the gas during dielectronic recombination of He+ is the w = 2 excitation energy of He+.
mean_kinetic_energy = 1.036e-16 * T
gasphase_recombination_cooling = {
    "H+": mean_kinetic_energy * gasphase_recombination_rates["H+"],
    "He+": 1.55e-26 * T**-0.3647,
    "He++": mean_kinetic_energy * gasphase_recombination_rates["He++"],
}
gasphase_recombination_cooling["He++"] = 4 * gasphase_recombination_cooling["H+"]  # H-like



"""Implementation of base Process class with methods for managing and solving systems of equations"""

from .symbols import n_, d_dt
from .equation import Equation
from .equation_system import EquationSystem


class Process:
    """
    Top-level class containing a description of a microscopic process

    Most importantly, this implements the procedure for combining processes to build up a network for chemistry
    + conservation equations.
    """

    def __init__(self, name="", bibliography={}):
        """Construct an empty Process instance

        Parameters
        ----------
        name: str, optional
            Name of the process
        """
        self.name = name
        self.initialize_network()
        self.rate = 0
        self.heat = 0
        self.bibliography = bibliography
        self.subprocesses = [self]

    def __repr__(self):
        """Print the name in print()"""
        return self.name

    def __add__(self, other):
        """Sum 2 processes together: define a new process whose rates are the sum of the input process"""
        if other == 0:  # necessary for native sum() routine to work
            return self

        attrs_to_sum = "heat", "subprocesses", "network"  # all rates

        sum_process = Process()
        sum_process.rate = None  # "rate" ceases to be meaningful for composite processes
        for summed_attr in attrs_to_sum:
            attr1, attr2 = getattr(self, summed_attr), getattr(other, summed_attr)
            if attr1 is None or attr2 is None:
                setattr(sum_process, summed_attr, None)
            else:
                setattr(sum_process, summed_attr, attr1 + attr2)

        sum_process.name = f"{self.name} + {other.name}"
        return sum_process

    def __radd__(self, other):
        return self.__add__(other)

    def initialize_network(self):
        self.network = EquationSystem()  # this is a dict for which unknown keys are initialized to 0 by default

    @property
    def heat(self):
        """Energy lost from gas per unit volume and time"""
        return self.__heat

    @heat.setter
    def heat(self, value):
        """Ensures that the network is always updated when we update the heat"""
        self.__heat = value
        self.network["heat"] = Equation(d_dt(n_("heat")), value)

    def solve(
        self,
        known_quantities,
        guess,
        time_dependent=[],
        dt=None,
        verbose=False,
        tol=1e-3,
        careful_steps=10,
    ):
        """
        Solves the equations for a set of desired quantities given a set of known quantities

        Parameters
        ----------
        known_quantities: dict
            Dict of symbolic quantities and their values that will be plugged into the network solve as known quantities.
            Can be arrays if you want to substitute multiple values. If T is included here, we solve for chemical
            equilibrium. If T is not included, solve for thermochemical equilibrium.
        guess: dict, optional
            Dict of symbolic quantities and their values that will be plugged into the network solve as guesses for the
            unknown quantities. Can be arrays if you want to substitute multiple values. Will default to trying sensible
            guesses for recognized quantities.
        normalize_to_H: bool, optional
            Whether to return abundances normalized by the number density of H nucleons (default: True)
        reduce_network: bool, optional
            Whether to solve the reduced version of the network substituting conservation laws (default: True)
        tol: float, optional
            Desired relative error in chemical abundances (default: 1e-3)
        careful_steps: int, optional
            Number of careful initial steps in the Newton solve before full step size is used - try increasing this if
            your solve has trouble converging.

        Returns
        -------
        soldict: dict
            Dict of solved quantities
        """

        return self.network.solve(
            known_quantities,
            guess,
            time_dependent=time_dependent,
            tol=tol,
            careful_steps=careful_steps,
            dt=dt,
            verbose=verbose,
        )

    def solver_functions(self, solve_vars, time_dependent=[], return_jac=False, return_dict=False):
        """Returns the RHS of the system to solve and its Jacobian, applying simplifications"""
        return self.network.solver_functions(solve_vars, time_dependent, return_jac, return_dict)

    def generate_code(self, solve_vars, time_dependent=[], language="c", jac=True, cse=True):
        """Generates numerical code that implements the system RHS and/or Jacobian in the specified language."""
        return self.network.generate_code(solve_vars, time_dependent, language, jac, cse)



"""Implementation of EquationSystem for representing, manipulating, and constructing systems of conservation laws"""

import sympy as sp
from .symbols import d_dt, n_, x_, t, BDF, n_Htot, internal_energy
from .data import SolarAbundances
from jax import numpy as jnp
import numpy as np
from .numerics import newton_rootsolve
from astropy import units
from .equation import Equation
from sympy.codegen.ast import Assignment


class EquationSystem(dict):
    """Dict of symbolic expressions with certain superpowers for manipulating sets of conservation equations."""

    def copy(self):
        new = EquationSystem()
        for k in self:
            new[k] = self[k]
        return new

    def __getitem__(self, __key: str):
        """Dict getitem method where we initialize a differential equation for the conservation of a species if the key
        does not exist"""
        if __key not in self:
            self.__setitem__(__key, Equation(d_dt(n_(__key)), 0))  # technically should only be n_ if this is a species
            # need to make sure that d/dt's don't add up when composing equations
        return super().__getitem__(__key)

    def __add__(self, other):
        """Return a dict whose values are the sum of the values of the operands"""
        keys = self.keys() | other.keys()
        new = EquationSystem()
        for k in keys:
            new[k] = self[k] + other[k]
        return new

    @property
    def symbols(self):
        """Returns the set of all symbols in the equations"""
        all = set()
        for e in self.values():
            all.update(e.free_symbols)
        if t in all:  # leave time out
            all.remove(t)
        return all

    @property
    def jacobian(self):
        """Returns a dict of dicts representing the Jacobian of the RHS of the system. Keys are the names of the
        conserved quantities and subkeys are the variable of differentiation.
        """
        return {k: {s: sp.diff(e.rhs, s) for s in self.symbols} for k, e in self.items()}

    def subs(self, expr, replacement):
        """Substitute symbolic expressions throughout the whole network."""
        for k, e in self.items():
            self[k] = e.subs(expr, replacement)

    def reduced(self, knowns, time_dependent=[]):
        subsystem = self.copy()
        subsystem.set_time_dependence(time_dependent)
        subsystem.do_conservation_reductions(time_dependent)
        if "T" in (str(k) for k in knowns) and "T" not in time_dependent:
            del subsystem["heat"]
        return subsystem

    def set_time_dependence(self, time_dependent_vars):
        """Insert backward-difference formulae or set to steady state"""
        # put in backward differences
        for q in self:
            if q in time_dependent_vars:  # insert backward-difference formula
                self[q] = Equation(BDF(q), self[q].rhs)
            else:
                self[q] = Equation(0, self[q].rhs)
        if "T" in time_dependent_vars:  # special behaviour
            self["heat"] = Equation(BDF("T"), self["heat"].rhs)
            if "u" not in self:
                self["u"] = Equation(0, sp.Symbol("u") - internal_energy)

    def do_conservation_reductions(self, time_dependent_vars):
        """Eliminate equations from the system using known conservation laws."""
        self.substitutions = []

        # since we have n_Htot let's convert all other n's to x's
        for s in self.symbols:
            if "n_" in str(s) and "Htot" not in str(s):
                species = str(s).split("_")[1]
                self.substitutions.append((s, n_Htot * x_(species)))

        # charge neutrality
        if "e-" not in time_dependent_vars:
            self.substitutions.append((x_("e-"), x_("H+") + x_("He+") + 2 * x_("He++")))
            del self["e-"]

        #  general: sum(n_(species containing H) / (number of H in species))  - n_("H_2") / 2 #
        if "H+" not in time_dependent_vars:
            self.substitutions.append((x_("H+"), 1 - x_("H")))
            if "H+" in self:
                del self["H+"]

        if "He++" not in time_dependent_vars:
            y = sp.Symbol("y")
            self.substitutions.append((x_("He++"), y - x_("He") - x_("He+")))
            if "He++" in self:
                del self["He++"]

        for expr, sub in self.substitutions:
            self.subs(expr, sub)

            # general: substitute highest ionization state with n_Htot * x_element - sum of lower ionization states

    @property
    def rhs(self):
        """Return as dict of rhs-lhs instead of equations"""
        return {k: e.rhs - e.lhs for k, e in self.items()}

    @property
    def rhs_scaled(self):
        """Returns a scaled version of the the RHS pulling out the usual factors affecting collision rates"""
        return [r for r in self.rhs.values()]  # / (T**0.5 * n_Htot * n_Htot * 1e-12)

    def solve(
        self,
        knowns,
        guesses,
        time_dependent=[],
        dt=None,
        verbose=False,
        tol=1e-3,
        careful_steps=10,
        symbolic_keys=False,
    ):
        """
        Solves for equilibrium after substituting a set of known quantities, e.g. temperature, metallicity,
        etc.

        Parameters
        ----------
        known_quantities: dict
            Dict of symbolic quantities and their values that will be plugged into the network solve as known quantities.
            Can be arrays if you want to substitute multiple values. If T is included here, we solve for chemical
            equilibrium. If T is not included, solve for thermochemical equilibrium.
        guesses: dict
            Dict of symbolic quantities and their values that will be plugged into the network solve as guesses for the
            unknown quantities. Can be arrays if you want to substitute multiple values. Will default to trying sensible
            guesses for recognized quantities (NOT IMPLEMENTED YET)
        tol: float, optional
            Desired relative error in chemical abundances (default: 1e-3)
        careful_steps: int, optional
            Number of careful initial steps in the Newton solve before full step size is used - try increasing this if
            your solve has trouble converging.

        Returns
        -------
        soldict: dict
            Dict of species and their equilibrium abundances relative to H or raw number densities (depending on
            value of normalize_to_H)
        """

        def printv(*a, **k):
            """Print only if locally verbose=True"""
            if verbose:
                print(*a, **k)

        # first: check knowns and guesses are all same size
        num_params = np.array([len(np.array(guesses[g])) for g in guesses] + [len(np.array(knowns[g])) for g in knowns])
        if not np.all(num_params == num_params[0]):
            raise ValueError("Input parameters and initial guesses must all have the same shape.")
        num_params = num_params[0]

        if dt is not None:
            knowns["Δt"] = np.repeat(dt.to(units.s), num_params)

        if "u" in guesses or "T" in time_dependent:
            self["u"] = Equation(0, internal_energy - sp.Symbol("u"))
        subsystem = self.reduced(knowns, time_dependent)
        symbols = subsystem.symbols
        num_equations = len(subsystem)

        # are there any symbols for which we can make a reasonable assumption or directly solve the steady-state approximation?
        prescriptions = {"y": SolarAbundances.x("He"), "Y": SolarAbundances.mass_fraction["He"], "Z": 1.0}
        assumed_values = {}
        if len(symbols) > num_equations + len(knowns):
            undetermined_symbols = symbols.difference(set(sp.Symbol(g) for g in guesses))
            printv(f"Undetermined symbols: {undetermined_symbols}")
            for s in undetermined_symbols:
                # if we have a prescription for this quantity, plug it in here. This should eventually be specified at the model level.
                if str(s) in prescriptions:
                    # case 1: we have given a value, which we should add to the list of knowns
                    assumed_values[str(s)] = np.repeat(prescriptions[str(s)], num_params)
                    printv(f"{s} not specified; assuming {s}={prescriptions[str(s)]}.")
                    symbols = subsystem.symbols
                    # case 2: we have given an expression in terms of the other available quantities: we need to subs it

        # ok now we should have number of symbols unknowns + knowns
        printv(
            f"Free symbols: {symbols}\nKnown values: {list(knowns)}\nAssumed values: {list(assumed_values)}\nEquations solved: {list(subsystem.rhs)}"
        )
        if len(symbols) != len(knowns | assumed_values) + len(subsystem):
            raise ValueError(
                f"Number of free symbols is {len(symbols)} != number of knowns {len(knowns)} + number of assumptions {len(assumed_values)} + number of equations {len(subsystem)}\n"
            )
        else:
            printv(
                f"It's solvin time. Solving for {set(guesses)} based on input {set(knowns)} and assumptions about {set(assumed_values)}"
            )

        guessvals = {}
        paramvals = {}
        for s in subsystem.symbols:
            for g in guesses:
                if g == str(s) or f"x_{g}" == str(s):
                    guessvals[s] = guesses[g]
            for k in knowns | assumed_values:
                if k == str(s) or f"x_{k}" == str(s):
                    paramvals[s] = (knowns | assumed_values)[k]

        lambda_args = [list(guessvals.keys()), list(paramvals.keys())]
        func = sp.lambdify(lambda_args, subsystem.rhs_scaled, modules="jax", cse=True)

        tolerance_vars = [x_("H"), x_("He+") + x_("He"), 1 - x_("H")]
        if "T" in guesses:
            tolerance_vars += [sp.Symbol("T")]
        if "u" in guesses:
            tolerance_vars += [sp.Symbol("u"), subsystem["heat"].rhs]
            # , subsystem["heat"]]  # converge on the internal energy and  cooling rate
        tolfunc = sp.lambdify(lambda_args, tolerance_vars, modules="jax", cse=True)

        def f_numerical(X, *params):
            """JAX function to rootfind"""
            return jnp.array(func(X, params))

        def tolerance_func(X, *params):
            """Solution will terminate if the relative change in this quantity is < tol"""
            return jnp.array(tolfunc(X, params))

        # option to bail here and just provide the RHS

        # jacfunc = sp.lambdify(
        #     lambda_args, [[sp.diff(a, g) for g in guessvals] for a in subsystem.rhs_scaled]
        # )  # , modules="jax", cse=True

        sol, num_iter = newton_rootsolve(
            f_numerical,
            jnp.array([g for g in guessvals.values()]).T,
            jnp.array([p for p in paramvals.values()]).T,
            tolfunc=tolerance_func,
            rtol=tol,
            careful_steps=careful_steps,
            nonnegative=True,
            return_num_iter=True,
        )

        soldict = self.package_solution(sol, guessvals, guesses, paramvals, subsystem, symbolic_keys)

        return soldict

    def package_solution(self, sol, guessvals, guesses, paramvals, subsystem, symbolic_keys):
        # now repack the solution
        soldict = {}
        for i, g in enumerate(guessvals):
            soldict[g] = sol[:, i]
        # do a reverse-pass on the substitutions we made to get all quantities
        values_to_subs = soldict | paramvals
        for expr, sub in reversed(subsystem.substitutions):
            if expr in soldict:
                continue
            if "n_" in str(expr):
                continue
            soldict[expr] = sp.lambdify(list(sub.free_symbols), sub)(
                *[values_to_subs[s] for s in list(sub.free_symbols)]
            )  # should probably make a function of this
            values_to_subs |= soldict
        if not symbolic_keys:
            soldict = {str(k): v for k, v in soldict.items()}
            # if we have a bunch of x_'s, should also link up keys in the original input format, e.g. H->x_H
            if np.any(["x_" in k for k in guesses]):  # if we specified abundances with x_ notation, return same
                return soldict
            soldict2 = {}  # otherwise return with input format where keys are simple species strings
            for k in soldict:
                if "x_" in str(k):
                    soldict2[str(k).replace("x_", "")] = soldict[k]
                else:
                    soldict2[k] = soldict[k]
            soldict = soldict2

        return soldict

    def solver_functions(self, solve_vars, time_dependent=[], return_jac=False, return_dict=False):
        """Returns the RHS of the system to solve and its Jacobian, applying simplifications"""

        solve_vars = list(solve_vars)
        if "u" in solve_vars or "T" in time_dependent:
            self["u"] = Equation(0, internal_energy - sp.Symbol("u"))
            solve_vars.append("u")

        knowns = self.symbols.difference(solve_vars)
        subsystem = self.reduced(knowns, time_dependent)

        rhs = {}
        for s in subsystem.symbols:
            for g in solve_vars:
                if str(s) == "T" and "T" in solve_vars:
                    rhs[s] = subsystem.rhs["heat"]
                elif str(g) == str(s) or f"x_{g}" == str(s):
                    rhs[s] = subsystem.rhs[g]

        if return_jac:
            jac = {}
            for s, expr in rhs.items():
                jac[s] = {s2: sp.diff(expr, s2) for s2 in rhs}

            if return_dict:
                return rhs, jac
            else:
                return (
                    list(rhs.values()),
                    [[jac[s1][s2] for s2 in rhs] for s1 in jac],
                    {s: i for i, s in enumerate(rhs)},
                )

        if return_dict:
            return rhs
        else:
            return list(rhs.values()), {s: i for i, s in enumerate(rhs)}

    def generate_code(self, solve_vars, time_dependent=[], language="Fortran", jac=True, cse=True, sanitize=True):
        """Generates numerical code that implements the system RHS and/or Jacobian in the specified language."""
        func, jac, indices = self.solver_functions(solve_vars, time_dependent, return_jac=jac)

        def printer(x, language="c"):
            match language.lower():
                case "fortran":
                    return sp.fcode(x, standard=2008)
                case "c":
                    return sp.ccode(x, standard="c99")
                case "python":
                    return sp.pycode(x)
                case "c++":
                    return sp.cxxcode(x, standard="c++11")

        codeblocks = []

        header = "# Computes the RHS function "
        if jac:
            header += "and Jacobian "
        header += f"to solve for {list(indices.keys())}\n\n"

        header += "# INDEX CONVENTION: " + " ".join(f"({i}: {s})" for s, i in indices.items())

        codeblocks.append(header)

        if cse:
            cse, (func, jac) = sp.cse((sp.Matrix(func), sp.Matrix(jac)))
            block = []
            for expr in cse:
                block.append(printer(Assignment(*expr), language))
            codeblocks.append(" \n".join(block))

        rhs_result = sp.MatrixSymbol("rhs_result", len(func), 1)
        codeblocks.append(printer(Assignment(rhs_result, func), language))

        if jac:
            jac_result = sp.MatrixSymbol("jac_result", len(func), len(func))
            codeblocks.append(printer(Assignment(jac_result, jac), language))

        code = "\n\n".join(codeblocks)
        if sanitize:
            sanitized_code = ""
            replacements = {"+": "plus"}
            for i, char in enumerate(code):
                if char in replacements:
                    if code[i - 1].split():  # if preceding character is not whitespace
                        sanitized_code += replacements[char]
                        continue
                sanitized_code += char
            code = sanitized_code
        return code



from .atomic_weights import atomic_weights


class SolarAbundancesClass:
    """Container for solar abundances with methods to convert between mass fraction and abundance per H"""

    bibliography = ["2009ARA&A..47..481A"]

    @property
    def mass_fraction(self):
        """Returns a hard-coded dict of Solar abundance mass fractions"""
        return {
            "Z": 0.0142,
            "He": 0.27030,
            "C": 2.53e-3,
            "N": 7.41e-4,
            "O": 6.13e-3,
            "Ne": 1.34e-3,
            "Mg": 7.57e-4,
            "Si": 7.12e-4,
            "S": 3.31e-4,
            "Ca": 6.87e-5,
            "Fe": 1.38e-3,
        }

    @property
    def abundance_per_H(self):
        """Returns dictionary of abundances per H nucleon"""
        return {species: f / (1 - f) / atomic_weights[species] for species, f in self.mass_fraction.items()}

    def x(self, species):
        return self.get_abundance(species)

    def get_mass_fraction(self, species: str) -> float:
        """Returns the mass fraction of a given species"""
        if species in self.mass_fraction:
            return self.mass_fraction[species]
        else:
            raise NotImplementedError(f"Solar abundance of {species} not available.")

    def get_abundance(self, species: str) -> float:
        """Returns the abundance per H nuclear of an input species"""
        if species in self.mass_fraction:
            return self.abundance_per_H[species]
        else:
            raise NotImplementedError(f"Solar abundance of {species} not available.")


SolarAbundances = SolarAbundancesClass()



"""Class specifying an N-body collisional process with generic methods"""

from ..process import Process
from ..symbols import n_
import sympy as sp


class NBodyProcess(Process):
    """Process implementing special methods specific to n-body processes, whose rates
    all follow the pattern

    rate per volume = k * prod_i(n_i) for i in species

    rate and heat are promoted from attributes to properties implemented to compute this pattern.

    Parameters
    ----------
    colliding_species:
        iterable of strings representing the colliding species.
    rate_coefficient: sympy.core.symbol.Symbol, optional
        Symbol symbol expression for k
    heat_rate_coefficient: sympy.core.symbol.Symbol, optional
        Symbol symbol expression for the heat rate coefficient = average radiated energy * k
    name: str, optional
        Name of the process
    """

    def __init__(self, colliding_species, rate_coefficient=0, heat_rate_coefficient=0, name: str = ""):
        self.name = name
        self.initialize_network()
        self.colliding_species = colliding_species
        self.rate_coefficient = rate_coefficient
        self.heat_rate_coefficient = heat_rate_coefficient
        self.subprocesses = [self]

    @property
    def heat_rate_coefficient(self):
        """Returns the heat rate coefficient of the N-body process"""
        return self.__heat_rate_coefficient

    @heat_rate_coefficient.setter
    def heat_rate_coefficient(self, value):
        """Ensures that the network is always updated when we update the rate coefficient"""
        self.__heat_rate_coefficient = value
        self.heat = value * sp.prod([n_(c) for c in self.colliding_species])

    @property
    def rate(self):
        """Returns the number of events per unit time and volume"""
        if self.rate_coefficient is None:
            return None
        return self.rate_coefficient * sp.prod([n_(c) for c in self.colliding_species])

    # @property
    # def heat(self):
    #     """Returns the energy radiated per unit time and volume"""
    #     return super().heat

    # #        self.__heat = self.heat_rate_coefficient * sp.prod([sp.Symbol("n_" + c) for c in self.colliding_species])
    # #        return self.__heat

    # @heat.setter
    # def heat(self, value):
    #     raise NotImplementedError(
    #         "Cannot directly set the heat value of an N-body process - set the rate coefficient instead."
    #     )

    @property
    def num_colliding_species(self):
        """Number of colliding species"""
        return len(self.colliding_species)



"""Specifies class for the local thermal and chemical state"""


class State:
    """Internal energy, abundances, dust mass fraction, radiation field, cosmic ray ionization rate,
    ortho-para ratio,"""

    f_ortho: float = 0.75
    hydrogen_massfrac: float = 0.7381
    metallicity: float = 0.0134

    def __init__(self):
        pass

    @property
    def density(self, species):
        return 0


#   @property
#    def



import jax, jax.numpy as jnp


def newton_rootsolve(
    func,
    guesses,
    params=[],
    jacfunc=None,
    tolfunc=None,
    rtol=1e-6,
    max_iter=100,
    careful_steps=1,
    nonnegative=False,
    return_num_iter=False,
):
    """
    Solve the system f(X,p) = 0 for X, where both f and X can be vectors of arbitrary length and p is a set of fixed
    parameters passed to f. Broadcasts and parallelizes over an arbitrary number of initial guesses and parameter
    choices.

    Parameters
    ----------
    func: callable
        A JAX function of signature f(X,params) that implements the function we wish to rootfind, where X and params
        are arrays of shape (n,) and (n_p,) for dimension n and parameter number n_p. In general can return an array of
        shape (m,)
    guesses: array_like
        Shape (n,) or (N,n) array_like where N is the number of guesses + corresponding parameter choices
    params: array_like
        Shape (n,) or (N,n_p) array_like where N is the number of guesses + corresponding parameter choices
    jacfunc: callable, optional
        Function with the same signature as f that returns the Jacobian of f - will be computed with autodiff from f if
        not specified.
    rtol: float, optional
        Relative tolerance - iteration will terminate if relative change in all quantities is less than this value.
    atol: float, optional
        Absolute tolerance: iteration will terminate if the value computed by tolfunc goes below this value.
    careful_steps: int, optional
        Number of "careful" initial steps to take, gradually ramping up the step size in the Newton iteration

    Returns
    -------
    X: array_like
        Shape (N,n) array of solutions
    """
    BIG, SMALL = 1e37, 1e-37

    guesses = jnp.array(guesses)
    guesses = jnp.where(nonnegative, guesses.clip(SMALL), guesses)
    params = jnp.array(params)
    if len(guesses.shape) < 2:
        guesses = jnp.atleast_2d(guesses).T
    if len(params.shape) < 2:
        params = jnp.atleast_2d(params).T

    if jacfunc is None:
        jac = jax.jacfwd(func)

    if tolfunc is None:

        def tolfunc(X, *params):
            return X

    def solve(guess, params):
        """Function to be called in parallel that solves the root problem for one guess and set of parameters"""

        def iter_condition(arg):
            """Iteration condition for the while loop: check if we are within desired tolerance."""
            X, dx, num_iter = arg
            fac = jnp.min(jnp.array([(num_iter + 1.0) / careful_steps, 1.0]))
            tol2, tol1 = tolfunc(X, *params), tolfunc(X - dx, *params)
            tolcheck = jnp.any(jnp.abs(tol1 - tol2) > rtol * jnp.abs(tol1) * fac)
            return jnp.any(jnp.abs(dx) > fac * rtol * jnp.abs(X)) & (num_iter < max_iter) & tolcheck

        def X_new(arg):
            """Returns the next Newton iterate and the difference from previous guess."""
            X, _, num_iter = arg
            fac = jnp.min(jnp.array([(num_iter + 1.0) / careful_steps, 1.0]))
            J = jac(X, *params)
            #  condition number is nice but possibly very slow due to batching, e.g. https://github.com/jax-ml/jax/issues/11321
            # there is no reason for this from a pure FLOPS standpoint!
            #            cond = jsp.linalg.cond(J)  # , p=2)
            #            dx = jnp.where(cond < 1e30, -jnp.linalg.solve(J, func(X, *params)) * fac, jnp.zeros_like(X))
            dx = -jnp.linalg.solve(J, func(X, *params)) * fac
            dx_finite = jnp.all(jnp.isfinite(dx))
            Xnew = jnp.where(dx_finite, jnp.where(nonnegative, (X + dx).clip(SMALL), X + dx), X)
            return Xnew, dx, num_iter + 1

        init_val = guess, 100 * guess, 0
        X, _, num_iter = jax.lax.while_loop(iter_condition, X_new, init_val)

        return X, num_iter

    X, num_iter = jax.vmap(solve)(guesses, params)
    if return_num_iter:
        return X, num_iter
    else:
        return X



from jaco.processes import (
    FreeFreeEmission,
    LineCoolingSimple,
    CollisionalIonization,
    GasPhaseRecombination,
    Ionization,
)
import numpy as np
from matplotlib import pyplot as plt
import sympy as sp
from jaco.data import SolarAbundances
import pytest
import matplotlib

matplotlib.use("Agg")

test_temperatures = [
    100.0,
]  # initial temperature guesses. We aspire to make the solver converge from any initial guess, but for now it's sensitive...


@pytest.mark.parametrize("T0", test_temperatures)
def test_neutral_cooling(T0):
    """Solve for thermochemical structure of neutral ISM at a range of densities. This can potentially change
    as we update the data or process implementations...
    """

    processes = (
        [CollisionalIonization(s) for s in ("H", "He", "He+")]
        + [GasPhaseRecombination(i) for i in ("H+", "He+", "He++")]
        + [FreeFreeEmission(i) for i in ("H+", "He+", "He++")]
        + [LineCoolingSimple(i) for i in ("H", "He+")]
    )
    system = sum(processes)

    # throw in some order-of-magntiude numbers to model photoelectric heating and cosmic-ray ionization
    heat_per_H = 1e-27
    zeta_CR = 2e-16
    system.heat += heat_per_H * sp.Symbol("n_Htot")
    system += Ionization(species="H", rate_per_volume=zeta_CR * sp.Symbol("n_Htot"))
    T = sp.Symbol("T")

    # C+ cooling - assumes x_C+ = x_e- = 3e-4
    x_C = 3e-4
    system.heat -= (
        1e-27
        * sp.Symbol("n_Htot")
        * sp.exp(-91.211 / T)
        * (4890 / sp.sqrt(T) * (x_C * sp.Symbol("n_Htot")) + 0.47 * T**0.15 * sp.Symbol("n_Htot"))
    )

    ngrid = np.logspace(-2, 3, 10**4)
    ones = np.ones_like(ngrid)

    knowns = {"n_Htot": ngrid}
    y = SolarAbundances.x("He")
    guesses = {
        "T": T0 * ones,
        "H": ones * 0.5,
        "He": y * ones * 0.99,
        "He+": y * ones * 0.01,
    }

    sol = system.solve(knowns, guesses, tol=1e-3)  # , careful_steps=30)

    fig, ax = plt.subplots(figsize=(3, 3))
    T_test = np.load("tests/neutral_cooling_testdata.npy")[:, 1]
    ax.loglog(ngrid, sol["T"], label=r"T (K)", color="black")
    ax.loglog(ngrid, T_test, label=r"T (K) (reference.)", color="red", ls="dotted")
    ax.loglog(ngrid, sol["H"], label=r"$x_{\rm H0}$", ls="dotted", color="black")
    ax.loglog(ngrid, sol["e-"], label=r"$x_{e-}$", ls="dashed", color="black")
    ax.set_xlabel(r"$n_{\rm H}\,\left(\rm cm^{-3}\right)$")
    ax.legend(labelspacing=0)
    ax.set_xlim(ngrid[0], ngrid[-1])
    ax.set_ylim(1e-4, 3e4)
    ax.set_yticks(10.0 ** np.arange(-4, 5))
    plt.savefig("tests/neutral_cooling.png", bbox_inches="tight")
    np.save("neutral_cooling_testdata.npy", np.c_[ngrid, sol["T"]])
    assert np.all(np.abs((T_test - sol["T"]) / sol["T"]) < 0.1)


if __name__ == "__main__":
    for t in test_temperatures:
        test_neutral_cooling(t)



from ..process import *
from .nbody_process import *
from .thermal_process import *
from .freefree_emission import *
from .line_cooling import *
from .ionization import *
from .recombination import *



"""Implementation of ionization process"""

from ..process import Process
from ..misc import ionize
from ..symbols import T, T5, T3, T6, n_e, n_
import sympy as sp
from astropy import units as u


class Ionization(Process):
    """
    Class describing an ionization process. Could be collisional, photo, or cosmic ray-induced.

    Implements method for setting the chemistry network terms
    """

    def __init__(self, species: str, rate_per_volume=0):
        self.species = species
        self.ionized_species = ionize(species)
        self.__ionization_energy = None
        super().__init__()
        self.__rate_per_volume = rate_per_volume
        self.update_network()

    @property
    def rate(self):
        return self.__rate_per_volume

    @rate.setter
    def rate(self, value):
        self.__rate_per_volume = value
        self.update_network()

    @property
    def ionization_energy(self):
        if self.__ionization_energy is None:
            self.__ionization_energy = ionization_energy(self.species)
        return self.__ionization_energy

    def update_network(self):
        """Sets up rate terms in the associated chemistry network for each species involved"""
        if self.rate is None:
            return
        self.network[self.species] -= self.rate
        self.network[self.ionized_species] += self.rate
        self.network["e-"] += self.rate


def ionization_energy(species, unit=u.erg):
    """Return the energy in erg required to ionize a species"""
    # NOTE: come back and get this from a proper datafile
    energies_eV = {"H": 13.6, "He": 24.59, "He+": 54.42}
    return energies_eV[species] * u.eV.to(unit)


collisional_ionization_cooling_rates = {
    "H": 1.27e-21 * sp.sqrt(T) * sp.exp(-157809.1 / T) / (1 + sp.sqrt(T5)),  # 1996ApJS..105...19K
    "He": 9.38e-22 * sp.sqrt(T) * sp.exp(-285335.4 / T) / (1 + sp.sqrt(T5)),  # 1996ApJS..105...19K
    "He+": 4.95e-22 * sp.sqrt(T) * sp.exp(-631515 / T) / (1 + sp.sqrt(T5)),  # 1996ApJS..105...19K
}

collisional_ionization_rates = {
    "H": 5.85e-11 * sp.sqrt(T) * sp.exp(-157809.1 / T) / (1 + sp.sqrt(T5)),  # 1996ApJS..105...19K
    "He": 2.38e-11 * sp.sqrt(T) * sp.exp(-285335.4 / T) / (1 + sp.sqrt(T5)),  # 1996ApJS..105...19K
    "He+": 5.68e-12 * sp.sqrt(T) * sp.exp(-631515 / T) / (1 + sp.sqrt(T5)),  # 1996ApJS..105...19K
}


def CollisionalIonization(species=None) -> Ionization:
    """Return an ionization process representing collisional ionization of the input species.

    Parameters
    ----------
    species: str, optional
        Species being collisionally ionized. If None, we compose all collisional ionization processes for all ions rates are known.

    Returns
    -------
    process: Ionization
        `Ionization` instance describing the collisional ionization process
    """

    if species is None:
        return sum([CollisionalIonization(s) for s in collisional_ionization_rates], Process())

    process = Ionization(species)
    process.name = f"Collisional Ionization of {species}"
    nprod = n_(species) * n_e

    if species not in collisional_ionization_rates:
        raise NotImplementedError(f"{species} does not have an available collisional ionization coefficient.")
    process.rate = collisional_ionization_rates[species] * nprod
    process.heat = -process.ionization_energy * process.rate

    return process



"""Class describing a generic heating/cooling process with no associated radiation or chemistry"""

from ..process import Process
from ..symbols import c_s, G, ρ, T, n_e, z
import sympy as sp


class ThermalProcess(Process):
    """Generic heating/cooling process"""

    def __init__(self, heating_rate, name=""):
        super().__init__(name=name)
        self.heat = heating_rate


PdV_heating = ThermalProcess(
    sp.Symbol("C_1") * c_s**2 * sp.sqrt(4 * sp.pi * G * ρ), name="Grav. Compression"
)  # 1998ApJ...495..346M

inv_compton_cooling = ThermalProcess(
    5.41e-36 * n_e * T * (1 + z) ** 4, name="Inverse Compton Cooling"
)  # 1986ApJ...301..522I



"""Implementation of free-free emission as a 2-body process"""

from .nbody_process import NBodyProcess
from ..misc import species_charge
from ..symbols import T
import sympy as sp


def gaunt_factor(T):
    """Fit to data in table 3.3 of Spitzer (1978)"""
    return 1.1 + 0.34 * sp.exp(-((5.5 - sp.log(T) / sp.log(10)) ** 2) / 3.0)  # 1996ApJS..105...19K


def FreeFreeEmission(ion: str) -> NBodyProcess:
    """Returns a free-free emisison process (i.e. bremmsstrahlung) for the input ion

    Parameters
    ----------
    ion: str
        Ion species

    Returns
    -------
    process: NBodyProcess
        `NBodyProcess` instance describing the cooling process
    """
    process = NBodyProcess({ion, "e-"})
    charge = species_charge(ion)
    if charge <= 0:
        raise ValueError(f"{ion} does not appear to be a cation - cannot do bremmstrahlung.")
    process.heat_rate_coefficient = -1.42e-27 * gaunt_factor(T) * charge**2 * sp.sqrt(T)  # 1996ApJS..105...19K
    return process



import numpy as np
import jax, jax.numpy as jnp
from ..solvers import newton_rootsolve


def test_newton_rootsolve_linear(N=10**5, dim=10):
    """Test: solve a linear system of dim variables with the Newton solver and check the solution.
    Solver should get the right answer to machine precision.
    """

    A = np.random.rand(N, dim, dim)
    cond = np.linalg.cond(A)
    A, cond = A[cond < 1e3], cond[cond < 1e3]  # just take the easy ones, we don't need to be fancy here
    N = len(A)
    b = np.random.rand(N, dim, 1)
    sol = np.linalg.solve(A, b)[:, :, 0]
    b = b[:, :, 0]

    @jax.jit
    def func(x, *params):
        """horrid function for mapping X, *params signature to linear residual Ax-b"""
        dim = x.shape[0]
        A = jnp.array(params)[: dim * dim].reshape((dim, dim))
        b = jnp.array(params[dim * dim : dim * (dim + 1)])
        return jnp.matmul(A, x) - b

    p = np.c_[A.reshape(N, dim * dim), b.reshape(N, dim)]
    guesses = np.copy(b)
    sol_newton = newton_rootsolve(func, guesses, p)  # [:,:,None]

    error = np.sum((sol_newton - sol) ** 2, axis=1) ** 0.5
    norm = np.sum(sol**2, axis=1)
    assert np.all(error < 1e-2 * norm)
